//
// Base64Encoding.cpp
//

#include <windows.h>
#include "Base64Encoding.h"

//
// needed data tables
//

static char s_vcAlphabet[] =
{
'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q',
'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h',
'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y',
'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
};

//
// needed definitions
//

#define PADDING_CHARACTER		'='

//
// Base64Encoding::encodeA --> ANSI ENCODE method
//

void Base64Encoding::encodeA( char* pszOutput, const char* pszInput )
{
	// call the internel helper method to do the job
	EncodeByteString( pszOutput, (BYTE*) pszInput, ::strlen( pszInput ) );
}

//
// Base64Encoding::encodeW --> UNICODE ENCODE method
//

void Base64Encoding::encodeW( char* pszOutput, const WCHAR* pszInput )
{
	// call the internel helper method to do the job
	EncodeByteString( pszOutput, (BYTE*) pszInput, ::wcslen( pszInput ) * sizeof( WCHAR ) );
}

//
// Base64Encoding::decode --> character set independent DECODE method
//

void Base64Encoding::decode( BYTE* pbOutput, char* pszInput, size_t sTerminatingChrs )
{
	// decode the input string

	for ( int i=0, j=::strlen( pszInput ), pos=0; i<j; i+=4 )
	{
		char*		pcCurr = &pszInput[ i ];

		if ( pcCurr[ 2 ] == PADDING_CHARACTER ) // XYPP -> x00
		{
			int		x = Base64ChrToIndex( pcCurr[ 0 ] );
			int		y = Base64ChrToIndex( pcCurr[ 1 ] );

			pbOutput[ pos++ ] = ( x << 2 ) | ( y >> 4 );
		}
		else if ( pcCurr[ 3 ] == PADDING_CHARACTER ) // XYZP -> xy0
		{
			int		x = Base64ChrToIndex( pcCurr[ 0 ] );
			int		y = Base64ChrToIndex( pcCurr[ 1 ] );
			int		z = Base64ChrToIndex( pcCurr[ 2 ] );

			pbOutput[ pos++ ] = ( x << 2 ) | ( y >> 4 );
			pbOutput[ pos++ ] = ( ( y & 0xF ) << 4 ) | ( z >> 2 );
		}
		else // XYZW -> xyz
		{
			int		x = Base64ChrToIndex( pcCurr[ 0 ] );
			int		y = Base64ChrToIndex( pcCurr[ 1 ] );
			int		z = Base64ChrToIndex( pcCurr[ 2 ] );
			int		w = Base64ChrToIndex( pcCurr[ 3 ] );

			pbOutput[ pos++ ] = ( x << 2 ) | ( y >> 4 );
			pbOutput[ pos++ ] = ( ( y & 0xF ) << 4 ) | ( z >> 2 );
			pbOutput[ pos++ ] = ( ( z & 3 ) << 6 ) | w;
		}
	}

	// write the terminating characters at the end of the decoded string (if required)

	for ( unsigned int t=0; t<sTerminatingChrs; t++ )
		pbOutput[ pos++ ] = '\0';

	// return to the caller
	return;
}

//
// Base64Encoding::required_chrsA --> HELPER METHOD for calculating the number of required characters (ANSI version)
//

size_t Base64Encoding::required_chrsA( char* pszInput )
{
	// return to the caller
	return ( ::strlen( pszInput ) / 3 ) * 4 + ( ( ::strlen( pszInput ) % 3 ) ? 4 : 0 );
}

//
// Base64Encoding::required_chrsW --> HELPER METHOD for calculating the number of required characters (UNICODE version)
//

size_t Base64Encoding::required_chrsW( WCHAR* pszInput )
{
	// return to the caller
	return ( ( ::wcslen( pszInput ) * sizeof( WCHAR ) ) / 3 ) * 4 + ( ( ( ::wcslen( pszInput ) * sizeof( WCHAR ) ) % 3 ) ? 4 : 0 );
}

//
// Base64Encoding::EncodeByteString --> HELPER METHOD furnishing the real implementation of the Base64 encoder
//
// ### REMEMBER THE TERMINATING NULL CHARACTER THAT IS PUT BY THIS ROUTINE! ###
//

void Base64Encoding::EncodeByteString( char* pszOutput, BYTE* pbInput, size_t sInputDim )
{
	// encode the input string

	for ( unsigned int i=0, pos=0; i<sInputDim; i+=3 )
	{
		BYTE*		pbCurr = &pbInput[ i ];

		if ( sInputDim - i == 1 ) // X00
		{
			pszOutput[ pos++ ] = s_vcAlphabet[ pbCurr[ 0 ] >> 2 ];
			pszOutput[ pos++ ] = s_vcAlphabet[ ( pbCurr[ 0 ] & 3 ) << 4 ];
			pszOutput[ pos++ ] = PADDING_CHARACTER;
			pszOutput[ pos++ ] = PADDING_CHARACTER;
		}
		else if ( sInputDim - i == 2 ) // XY0
		{
			pszOutput[ pos++ ] = s_vcAlphabet[ pbCurr[ 0 ] >> 2 ];
			pszOutput[ pos++ ] = s_vcAlphabet[ ( ( pbCurr[ 0 ] & 3 ) << 4 ) | ( pbCurr[ 1 ] >> 4 ) ];
			pszOutput[ pos++ ] = s_vcAlphabet[ ( pbCurr[ 1 ] & 0xF ) << 2 ];
			pszOutput[ pos++ ] = PADDING_CHARACTER;
		}
		else // XYZ
		{
			pszOutput[ pos++ ] = s_vcAlphabet[ pbCurr[ 0 ] >> 2 ];
			pszOutput[ pos++ ] = s_vcAlphabet[ ( ( pbCurr[ 0 ] & 3 ) << 4 ) | ( pbCurr[ 1 ] >> 4 ) ];
			pszOutput[ pos++ ] = s_vcAlphabet[ ( ( pbCurr[ 1 ] & 0xF ) << 2 ) | ( pbCurr[ 2 ] >> 6 ) ];
			pszOutput[ pos++ ] = s_vcAlphabet[ pbCurr[ 2 ] & 0x3F ];
		}
	}

	pszOutput[ pos ] = '\0';

	// return to the caller
	return;
}

//
// Base64Encoding::Base64ChrToIndex --> HELPER METHOD for obtaining the characted index in Base64
//

int Base64Encoding::Base64ChrToIndex( char cBase64Chr )
{
	for( int i=0; i<64; i++ )
		if ( cBase64Chr == s_vcAlphabet[ i ] )
			return i;
	return -1;
}
