//
// DllMain.cpp
//

#include <windows.h>
#include <stdio.h>

#include "..\include\WebServerCommon.h"
#include "Base64Encoding.h"

//
// EXTENSION ENTRY POINT
//

BOOL WINAPI DllMain(
  HINSTANCE hinstDLL,  // handle to the DLL module
  DWORD fdwReason,     // reason for calling function
  LPVOID lpvReserved   // reserved
)
{
	// initializes or uninitializes global resources

	if ( fdwReason == DLL_PROCESS_ATTACH )
	{
	}
	else if ( fdwReason == DLL_PROCESS_DETACH )
	{
	}

	return TRUE;
}

//
// Definitions.
//

#define ROOTXML "<Tree TreeId=\"console\" NodeXmlSrc=\"/default/MsdnTreeXmlGen.dyn\"><TreeNode NodeId=\"mycomputer\" Title=\"%s\" NodeXmlSrc=\"/default/MsdnTreeXmlGen.dyn?MyComputer\" Href=\"/default/MsdnTreeXmlGen.dyn?RightFrameStart\" NodeImgSrc=\"/default/gif/MyComputer.gif\"></TreeNode></Tree>"

#define NODEXML_START			"<TreeNode NodeId=\"%s\" Title=\"%s\""
#define NODEXML_PARENTXMLSRC	"ParentXmlSrc=\"%s\""
#define NODEXML_NODEXMLSRC		"NodeXmlSrc=\"%s\""
#define NODEXML_HREF			"Href=\"%s\""
#define NODEXML_NODEIMGSRC		"NodeImgSrc=\"%s\""
#define NODEXML_END				"</TreeNode>"

#define MAX_BUFF_SIZE       (1024*1024)

//
// STL stuff
//

#include <vector>
#include <string>
typedef std::basic_string< CHAR > charstring;
typedef std::basic_string< WCHAR > widestring;
typedef std::vector< widestring > widestring_vector;

//
// Helper_ScanAndCorrectXmlContent --> helper function that searches for invalid Xml characters and if found replaces them with suitable ones
//

void Helper_ScanAndCorrectXmlContent( charstring& csContent )
{
	// do the operation

	for ( charstring::size_type i=0; i<csContent.size(); i++ )
	{
		unsigned char		c = csContent[ i ];
		if ( c == '&' || c == '"' || c >= 0x80 )
		{
			char		szHex[ 32 ];
			::sprintf( szHex, "&#x%x;", c );
			csContent.replace( i, 1, szHex );
			i += ::strlen( szHex ) - 1;
		}
	}

	// return to the caller

	return;
}

void Helper_ScanAndCorrectXmlContent( char* pszXmlContent, size_t sXmlContentBufferDim ) // NEW VERSION !!
{
	// do the operation

	charstring		csXmlContent = pszXmlContent;
	Helper_ScanAndCorrectXmlContent( csXmlContent );
	::strcpy( pszXmlContent, csXmlContent.substr( 0, sXmlContentBufferDim - 1 ).c_str() );

	// return to the caller
	return;
}

//
// Helper_StripFolderName --> helper function for striping the name of a folder from the folder full name
//

void Helper_StripFolderName( WCHAR* pszOutput, const WCHAR* pszInput )
{
	// strip the name

	if ( ::wcslen( pszInput ) <= 3 )
		::wcscpy( pszOutput, pszInput );
	else
	{
		// detect the eventual ending slash

		const WCHAR*		p = pszInput + ::wcslen( pszInput ) - 1;
		if ( *p == L'\\' || *p == L'/' )
			p--;

		for ( int i=0; *p != L'\\' && *p != L'/' && p >= pszInput; i++, p-- );

		// fill the output buffer

		::memcpy( pszOutput, p + 1, i * sizeof(WCHAR) );
		pszOutput[ i ] = L'\0';
	}

	// return to the caller
	return;
}

//
// Return_RootXml --> return to the client the default root xml
//

size_t Return_RootXml( UINT uiProtocolVersion, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen )
{
	size_t		sResponseLen = 0;

	// return the default xml for the root

	char		szComputerName[ MAX_COMPUTERNAME_LENGTH + 1 ];
	DWORD		dwComputerNameLen = sizeof( szComputerName );
	::GetComputerName( szComputerName, &dwComputerNameLen );

	char		szRootXml[ 1024 ];
	::sprintf( szRootXml, ROOTXML, szComputerName );

	SResponseCharacteristics		srchrs;
	ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
	srchrs.pszContentType = "text/xml";
	srchrs.uiContentLength = ::strlen( szRootXml );
	srchrs.pbData = (BYTE*) szRootXml;
	sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

	// return to the caller
	return sResponseLen;
}

//
// Return_RootNodeXml --> return to the client the xml of the root node
//

#define APPEND ::strcat( szRootNodeXml, szBuffer );

size_t Return_RootNodeXml( UINT uiProtocolVersion, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen )
{
	size_t		sResponseLen = 0;

	// return the xml for the root node

	char		szComputerName[ MAX_COMPUTERNAME_LENGTH + 1 ];
	DWORD		dwComputerNameLen = sizeof( szComputerName );
	::GetComputerName( szComputerName, &dwComputerNameLen );

	char		szRootNodeXml[ 16 * 1024 ];
	char		szBuffer[ 16 * 1024 ];

	::sprintf( szRootNodeXml, NODEXML_START, "mycomputer", szComputerName );
	::sprintf( szBuffer, " " NODEXML_PARENTXMLSRC, "/default/MsdnTreeXmlGen.dyn" ); APPEND
	::sprintf( szBuffer, " " NODEXML_NODEIMGSRC, "/default/gif/MyComputer.gif" ); APPEND
	::sprintf( szBuffer, " " NODEXML_NODEXMLSRC ">", "/default/MsdnTreeXmlGen.dyn?MyComputer" ); APPEND

		::sprintf( szBuffer, NODEXML_START, "local_drives", "Local Drives" ); APPEND
		::sprintf( szBuffer, " " NODEXML_NODEIMGSRC, "/default/gif/LocalDrives.gif" ); APPEND
		::sprintf( szBuffer, " " NODEXML_HREF, "/default/MsdnTreeXmlGen.dyn?LocalDrivesStart" ); APPEND
		::sprintf( szBuffer, " " NODEXML_NODEXMLSRC ">", "/default/MsdnTreeXmlGen.dyn?LocalDrives" ); APPEND
		::strcat( szRootNodeXml, NODEXML_END );

	::strcat( szRootNodeXml, NODEXML_END );

	// compose the response text and set the web server buffer

	SResponseCharacteristics		srchrs;
	ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
	srchrs.pszContentType = "text/xml";
	srchrs.uiContentLength = ::strlen( szRootNodeXml );
	srchrs.pbData = (BYTE*) szRootNodeXml;
	sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

	// return to the caller
	return sResponseLen;
}

#undef APPEND

//
// Return_LocalDrives --> return the xml for representing the local drives
//

#define APPEND ::strcat( szXml, szBuffer );

size_t Return_LocalDrives( UINT uiProtocolVersion, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen )
{
	size_t		sResponseLen = 0;

	// return the xml for the local drives

	char		szXml[ 16 * 1024 ];
	char		szBuffer[ 16 * 1024 ];

	::sprintf( szXml, NODEXML_START, "local_drives", "Local Drives" );
	::sprintf( szBuffer, " " NODEXML_PARENTXMLSRC, "/default/MsdnTreeXmlGen.dyn?MyComputer" ); APPEND
	::sprintf( szBuffer, " " NODEXML_NODEIMGSRC, "/default/gif/LocalDrives.gif" ); APPEND
	::sprintf( szBuffer, " " NODEXML_NODEXMLSRC ">", "/default/MsdnTreeXmlGen.dyn?LocalDrives" ); APPEND

	// get the logical drives in the system

	WCHAR		szLocalDrives[ 1024 ];
	::GetLogicalDriveStringsW( sizeof( szLocalDrives ) / sizeof( WCHAR ), szLocalDrives );

	for ( WCHAR* p = szLocalDrives; *p != L'\0'; p++ )
	{
		// isolate the drive root directory name

		WCHAR		szDriveName[ 32 ];
		for ( WCHAR* d = szDriveName; *p != L'\0' && d - szDriveName < ( sizeof( szDriveName ) / sizeof( WCHAR ) - 1 ); d++, p++ )
			*d = *p;
		*d = L'\0';

		// encode the drive name so it can be transferred with xml data

		char		szEncodedDriveName[ 32 * 2 ];
		Base64Encoding::encodeW( szEncodedDriveName, szDriveName );

		// compose the title for this drive

		UINT		uiDriveType = ::GetDriveTypeW( szDriveName );

		char		szDriveTitle[ MAX_PATH * 6 ]; // * 6 because of xml escapes

		WCHAR		szDriveW[ 3 ];
		::memcpy( szDriveW, szDriveName, 2 * sizeof( WCHAR ) );
		szDriveW[ 2 ] = L'\0';

		char		szDriveA[ 3 ];
		::WideCharToMultiByte( CP_ACP, 0, szDriveW, -1, szDriveA, sizeof( szDriveA ), NULL, NULL );

		WCHAR		szVolumeNameW[ MAX_PATH ];
		BOOL		bGVI = ::GetVolumeInformationW( szDriveName,
			szVolumeNameW, sizeof( szVolumeNameW ) / sizeof( WCHAR ),
			NULL, NULL, NULL,
			NULL, 0 );

		if ( bGVI == FALSE )
		{
			switch( uiDriveType )
			{
			case DRIVE_REMOVABLE:
				::sprintf( szDriveTitle, "Removable (%s)", szDriveA );
				break;
			case DRIVE_FIXED:
				::sprintf( szDriveTitle, "Hard disk (%s)", szDriveA );
				break;
			case DRIVE_CDROM:
				::sprintf( szDriveTitle, "Cd-Rom (%s)", szDriveA );
				break;
			default:
				::strcpy( szDriveTitle, szDriveA );
				break;
			}
		}
		else
		{
			char		szVolumeNameA[ MAX_PATH ];
			::WideCharToMultiByte( CP_ACP, 0, szVolumeNameW, -1, szVolumeNameA, sizeof( szVolumeNameA ), NULL, NULL );

			::sprintf( szDriveTitle, "%s (%s)", szVolumeNameA, szDriveA );
		}

		// eventually correct the drive title replacing invalid XML characters

		::Helper_ScanAndCorrectXmlContent( szDriveTitle, sizeof( szDriveTitle ) );

		// append the xml for this logical drive

		char*		pszNodeImgSrc;
		switch ( uiDriveType )
		{
		case DRIVE_REMOVABLE:
			pszNodeImgSrc = "/default/gif/DriveRemovable.gif";
			break;
		case DRIVE_FIXED:
			pszNodeImgSrc = "/default/gif/DriveFixed.gif";
			break;
		case DRIVE_CDROM:
			pszNodeImgSrc = "/default/gif/DriveCdRom.gif";
			break;
		default:
			pszNodeImgSrc = "/default/gif/DriveOther.gif";
			break;
		}

		char		szXmlSrc[ MAX_PATH ];
		::sprintf( szXmlSrc, "/default/MsdnTreeXmlGen.dyn?Browse%s", szEncodedDriveName );

		char		szHref[ MAX_PATH ];
		::sprintf( szHref, "/default/MsdnTreeXmlGen.dyn?FilePage%s", szEncodedDriveName );

		::sprintf( szBuffer, NODEXML_START, szEncodedDriveName, szDriveTitle ); APPEND
		::sprintf( szBuffer, " " NODEXML_NODEIMGSRC, pszNodeImgSrc ); APPEND
		::sprintf( szBuffer, " " NODEXML_HREF, szHref ); APPEND
		::sprintf( szBuffer, " " NODEXML_NODEXMLSRC ">", szXmlSrc ); APPEND
		::strcat( szXml, NODEXML_END );
	}

	::strcat( szXml, NODEXML_END );

	// compose the response text and set the web server buffer

	SResponseCharacteristics		srchrs;
	ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
	srchrs.pszContentType = "text/xml";
	srchrs.uiContentLength = ::strlen( szXml );
	srchrs.pbData = (BYTE*) szXml;
	sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

	// return to the caller
	return sResponseLen;
}

#undef APPEND

//
// Return_Folder --> return the xml representing the contents of a specified folder
//

#define APPEND ::strcat( szXml, szBuffer );

size_t Return_Folder( UINT uiProtocolVersion, char* pszFolderName, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen )
{
	size_t		sResponseLen = 0;

	// decode the folder name

	WCHAR		szFolderName[ MAX_PATH ];
	if ( ( ::strlen( pszFolderName ) / 4 ) * 3 >= sizeof( szFolderName ) - sizeof( WCHAR ) )
		return 0;
	Base64Encoding::decode( (BYTE*) szFolderName, pszFolderName, sizeof( WCHAR ) );

	// return the xml for the specified folder

	char		szParentXmlSrc[ MAX_PATH * 4 ] = "/default/MsdnTreeXmlGen.dyn?";
	if ( ::wcslen( szFolderName ) <= 3 )
		::strcat( szParentXmlSrc, "LocalDrives" );
	else
	{
		// get the name of the parent folder

		WCHAR		szParentName[ MAX_PATH ];
		::wcscpy( szParentName, szFolderName );
		for ( int i=::wcslen(szParentName)-1; i>=0; i-- )
			if ( szParentName[ i ] == L'\\' || szParentName[ i ] == L'/' )
			{
				if ( szParentName[ i+1 ] != L'\0' )
				{
					szParentName[ i+1 ] = L'\0';
					break;
				}
			}

		::strcat( szParentXmlSrc, "Browse" );

		char		szEncodedParentName[ MAX_PATH * 4 ];
		Base64Encoding::encodeW( szEncodedParentName, szParentName );

		::strcat( szParentXmlSrc, szEncodedParentName );
	}

	char		szNodeXmlSrc[ MAX_PATH * 4 ] = "/default/MsdnTreeXmlGen.dyn?Browse";
	::strcat( szNodeXmlSrc, pszFolderName );

	WCHAR		szTitleW[ MAX_PATH ];
	::Helper_StripFolderName( szTitleW, szFolderName );

	char		szTitleA[ MAX_PATH * 6 ]; // * 6 because of xml escapes
	::WideCharToMultiByte( CP_ACP, 0, szTitleW, -1, szTitleA, sizeof( szTitleA ), NULL, NULL );
	::Helper_ScanAndCorrectXmlContent( szTitleA, sizeof( szTitleA ) );

	char*		szXml = (char*) ::malloc( MAX_BUFF_SIZE );
	char		szBuffer[ 16 * 1024 ];

	::sprintf( szXml, NODEXML_START, pszFolderName, szTitleA );
	::sprintf( szBuffer, " " NODEXML_PARENTXMLSRC, szParentXmlSrc ); APPEND
	::sprintf( szBuffer, " " NODEXML_NODEXMLSRC ">", szNodeXmlSrc ); APPEND

	BOOL		bThereIsSomething = FALSE;

	// browse the specified folder and compose the xml

	HANDLE				hFindFirstFile = INVALID_HANDLE_VALUE;
	WCHAR				szFileName[ MAX_PATH ];

	::wcscpy( szFileName, szFolderName );
	WCHAR		cLastChr = szFileName[ ::wcslen( szFileName ) - 1 ];
	if ( cLastChr == L'\\' || cLastChr == L'/' )
		::wcscat( szFileName, L"*" );
	else
		::wcscat( szFileName, L"\\*" );

	int					iFindDataNum = 1024;
	int					iFindDataId = 0;
	WIN32_FIND_DATAW*	pwfdFindData = (WIN32_FIND_DATAW*) ::malloc( iFindDataNum * sizeof( WIN32_FIND_DATAW ) );

	while( TRUE )
	{
		// fill the pwfdFindData buffer with information regarding the specified pattern

		if ( hFindFirstFile == INVALID_HANDLE_VALUE )
		{
			hFindFirstFile = ::FindFirstFileW( szFileName, &pwfdFindData[ iFindDataId ] );
			if ( hFindFirstFile == INVALID_HANDLE_VALUE )
				break;
		}
		else
		{
			if ( ::FindNextFileW( hFindFirstFile, &pwfdFindData[ iFindDataId ] ) == FALSE )
				break;
		}

		// increment the iFindDataId counter and check whether a reallocation is needed

		iFindDataId++;
		if ( iFindDataId == iFindDataNum )
		{
			iFindDataNum += 1024;
			pwfdFindData = (WIN32_FIND_DATAW*) ::realloc( pwfdFindData, iFindDataNum * sizeof( WIN32_FIND_DATAW ) );
		}
	}

	// put the directories at the beginning of the list

	int		iFindDataCurr;

	int		iFindDataFirstFile = iFindDataId;

	for ( iFindDataCurr = 0; iFindDataCurr < iFindDataId; iFindDataCurr ++ )
	{
		WIN32_FIND_DATAW*		pwfdThis = &pwfdFindData[ iFindDataCurr ];

		if ( !( pwfdThis->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) )
		{
			BOOL		bDirectoryFound = FALSE;

			for( int iFindDataRemaining = iFindDataCurr + 1; iFindDataRemaining < iFindDataId; iFindDataRemaining ++ )
			{
				WIN32_FIND_DATAW*		pwfdRemaining = &pwfdFindData[ iFindDataRemaining ];

				if ( pwfdRemaining->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
				{
					WIN32_FIND_DATAW		wfdSwapHelper = *pwfdThis;
					*pwfdThis = *pwfdRemaining;
					*pwfdRemaining = wfdSwapHelper;

					bDirectoryFound = TRUE;
					break;
				}
			}

			if ( bDirectoryFound == FALSE )
			{
				iFindDataFirstFile = iFindDataCurr;
				break;
			}
		}
	}

	// order the lists alphabetically

	for ( iFindDataCurr = 0; iFindDataCurr < iFindDataFirstFile; iFindDataCurr ++ )
	{
		WIN32_FIND_DATAW*		pwfdThis = &pwfdFindData[ iFindDataCurr ];

		int					iFindDataPrevious = iFindDataCurr;

		for ( int iFindDataRemaining = iFindDataCurr + 1; iFindDataRemaining < iFindDataFirstFile; iFindDataRemaining ++ )
		{
			WIN32_FIND_DATAW*		pwfdRemaining = &pwfdFindData[ iFindDataRemaining ];

			if ( ::wcsicmp( pwfdFindData[ iFindDataPrevious ].cFileName, pwfdRemaining->cFileName ) > 0 )
				iFindDataPrevious = iFindDataRemaining;
		}

		if ( iFindDataPrevious != iFindDataCurr )
		{
			WIN32_FIND_DATAW*		pwfdPrevious = &pwfdFindData[ iFindDataPrevious ];

			WIN32_FIND_DATAW		wfdSwapHelper = *pwfdThis;
			*pwfdThis = *pwfdPrevious;
			*pwfdPrevious = wfdSwapHelper;
		}
	}

	for ( ; iFindDataCurr < iFindDataId; iFindDataCurr ++ )
	{
		WIN32_FIND_DATAW*		pwfdThis = &pwfdFindData[ iFindDataCurr ];

		int					iFindDataPrevious = iFindDataCurr;

		for ( int iFindDataRemaining = iFindDataCurr + 1; iFindDataRemaining < iFindDataId; iFindDataRemaining ++ )
		{
			WIN32_FIND_DATAW*		pwfdRemaining = &pwfdFindData[ iFindDataRemaining ];

			if ( ::wcsicmp( pwfdFindData[ iFindDataPrevious ].cFileName, pwfdRemaining->cFileName ) > 0 )
				iFindDataPrevious = iFindDataRemaining;
		}

		if ( iFindDataPrevious != iFindDataCurr )
		{
			WIN32_FIND_DATAW*		pwfdPrevious = &pwfdFindData[ iFindDataPrevious ];

			WIN32_FIND_DATAW		wfdSwapHelper = *pwfdThis;
			*pwfdThis = *pwfdPrevious;
			*pwfdPrevious = wfdSwapHelper;
		}
	}

	// browse the list of files

	for ( iFindDataCurr = 0; iFindDataCurr < iFindDataId; iFindDataCurr ++ )
	{
		WIN32_FIND_DATAW*		pwfdThis = &pwfdFindData[ iFindDataCurr ];

		// compose the full path of the file

		WCHAR		szFullPath[ MAX_PATH ];
		::wcscpy( szFullPath, szFileName );
		::wcscpy( szFullPath + ::wcslen( szFullPath ) - 1, pwfdThis->cFileName );

		// compose the xml

		BOOL		bCheckMemory = FALSE;

		if ( pwfdThis->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			if ( ::wcsicmp( pwfdThis->cFileName, L"." ) != 0 &&
				::wcsicmp( pwfdThis->cFileName, L".." ) != 0 )
			{
				char*		pszGifName = "/default/gif/Directory.gif";

				::wcscat( szFullPath, L"\\" );

				char		szEncodedFullPath[ MAX_PATH * 4 ];
				if ( Base64Encoding::required_chrsW( szFullPath ) < sizeof( szEncodedFullPath ) - 1 )
				{
					Base64Encoding::encodeW( szEncodedFullPath, szFullPath );

					char		szXmlSrc[ MAX_PATH * 5 ];
					::sprintf( szXmlSrc, "/default/MsdnTreeXmlGen.dyn?Browse%s", szEncodedFullPath );

					char		szTitle[ MAX_PATH * 6 ]; // * 6 because of xml escapes
					::WideCharToMultiByte( CP_ACP, 0, pwfdThis->cFileName, -1, szTitle, sizeof( szTitle ), NULL, NULL );
					::Helper_ScanAndCorrectXmlContent( szTitle, sizeof( szTitle ) );

					char		szHref[ MAX_PATH * 5 ];
					::sprintf( szHref, "/default/MsdnTreeXmlGen.dyn?FilePage%s", szEncodedFullPath );

					::sprintf( szBuffer, NODEXML_START, szEncodedFullPath, szTitle ); APPEND
					::sprintf( szBuffer, " " NODEXML_NODEIMGSRC, pszGifName ); APPEND
					::sprintf( szBuffer, " " NODEXML_HREF, szHref ); APPEND
					::sprintf( szBuffer, " " NODEXML_NODEXMLSRC ">", szXmlSrc ); APPEND
					::strcat( szXml, NODEXML_END );

					bThereIsSomething = TRUE;
					bCheckMemory = TRUE;
				}
			}
		}
		else
		{
			// compose the xml

			char		szEncodedFullPath[ MAX_PATH * 4 ];
			if ( Base64Encoding::required_chrsW( szFullPath ) < sizeof( szEncodedFullPath ) - 1 )
			{
				Base64Encoding::encodeW( szEncodedFullPath, szFullPath );

				char		szTitle[ MAX_PATH * 6 ]; // * 6 because of xml escapes
				::WideCharToMultiByte( CP_ACP, 0, pwfdThis->cFileName, -1, szTitle, sizeof( szTitle ), NULL, NULL );
				::Helper_ScanAndCorrectXmlContent( szTitle, sizeof( szTitle ) );

				char		szHref[ MAX_PATH * 5 ];
				::sprintf( szHref, "/default/MsdnTreeXmlGen.dyn?FilePage%s", szEncodedFullPath );

				char		szGifName[ MAX_PATH * 5 ];
				::sprintf( szGifName, "/default/MsdnTreeXmlGen.dyn?SmallIcon%s", szEncodedFullPath );

				::sprintf( szBuffer, NODEXML_START, szEncodedFullPath, szTitle ); APPEND
				::sprintf( szBuffer, " " NODEXML_HREF, szHref ); APPEND
				::sprintf( szBuffer, " " NODEXML_NODEIMGSRC ">", szGifName ); APPEND
				::strcat( szXml, NODEXML_END );

				bThereIsSomething = TRUE;
				bCheckMemory = TRUE;
			}
		}

		if ( bCheckMemory )
		{
			if ( ::strlen( szXml ) > MAX_BUFF_SIZE - 10 * 1024 )
				break;
		}
	}

	if ( hFindFirstFile != INVALID_HANDLE_VALUE )
		::FindClose( hFindFirstFile );

	// close the xml stream

	if ( bThereIsSomething == FALSE )
	{
		char		szNodeId[ MAX_PATH * 4 ] = "Nothing";
		::strcat( szNodeId, pszFolderName );

		::sprintf( szBuffer, NODEXML_START ">", szNodeId, "no files or directories" ); APPEND
		::strcat( szXml, NODEXML_END );
	}

	::strcat( szXml, NODEXML_END );

	// compose the response text and set the web server buffer

	if ( hFindFirstFile != INVALID_HANDLE_VALUE )
	{
		SResponseCharacteristics		srchrs;
		ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
		srchrs.pszContentType = "text/xml";
		srchrs.uiContentLength = ::strlen( szXml );
		srchrs.pbData = (BYTE*) szXml;
		sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );
	}
	else
	{
		sResponseLen = ssf.pfnReturn_InternalServerError500( uiProtocolVersion, pbResponse, sResponseBufferLen );
	}

	// return to the caller, freeing the allocated memory
	::free( szXml );
	::free( pwfdFindData );
	return sResponseLen;
}

#undef APPEND

//
// Return_Icon --> return the icon associated to the specified file
//

size_t Return_Icon( UINT uiProtocolVersion, char* pszFileName, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen, BOOL bSmallIcon )
{
	size_t		sResponseLen = 0;

	// decode the file name

	WCHAR		szFileName[ MAX_PATH ];
	if ( ( ::strlen( pszFileName ) / 4 ) * 3 >= sizeof( szFileName ) - sizeof( WCHAR ) )
		return 0;
	Base64Encoding::decode( (BYTE*) szFileName, pszFileName, sizeof( WCHAR ) );

	// call the Shell Api for getting the icon of the specified file

	SHFILEINFOW	shfiInfo;
	DWORD		dwSHGFIResult = ::SHGetFileInfoW( szFileName, 0,
		&shfiInfo, sizeof( shfiInfo ),
		SHGFI_ICON | ( bSmallIcon ? SHGFI_SMALLICON : SHGFI_LARGEICON ) );
	if ( dwSHGFIResult == 0 )
	{
		WCHAR		szGenericExeFullPath[ MAX_PATH ];
		if( !GetEnvironmentVariableW( L"SYSTEMROOT", szGenericExeFullPath, sizeof( szGenericExeFullPath ) / sizeof( WCHAR ) ) )
			return 0;
		::wcscat( szGenericExeFullPath, L"\\system32\\ntoskrnl.exe" );

		dwSHGFIResult = ::SHGetFileInfoW( szGenericExeFullPath, 0,
			&shfiInfo, sizeof( shfiInfo ),
			SHGFI_ICON | ( bSmallIcon ? SHGFI_SMALLICON : SHGFI_LARGEICON ) );
		if ( dwSHGFIResult == 0 )
			return 0;
	}

	// do the job of returning the large icon to the requesting client (as a small or a large icon)

	DWORD		dwIconDim = bSmallIcon ? 16 : 32;

	HDC			hDC = ::CreateCompatibleDC( NULL );
	HBITMAP		hBitmap = ::CreateBitmap( dwIconDim, dwIconDim, 1, 32, NULL );
	HBITMAP		hOldBitmap = (HBITMAP)::SelectObject( hDC, hBitmap );

	HBRUSH		hBackgBrush = ::CreateSolidBrush( RGB( 241, 241, 241 ) );
	RECT		rRect = { 0, 0, dwIconDim, dwIconDim };
	::FillRect( hDC, &rRect, hBackgBrush );
	::DeleteObject( hBackgBrush );

	::DrawIconEx( hDC, 0, 0, shfiInfo.hIcon, dwIconDim, dwIconDim, 0, NULL, DI_NORMAL );

	BITMAP		bmp;
	if ( ::GetObject( hBitmap, sizeof( BITMAP ), &bmp ) )
	{
		WORD			cClrBits = (WORD)(bmp.bmPlanes * bmp.bmBitsPixel);
		if ( cClrBits == 1 )
			cClrBits = 1;
		else if ( cClrBits <= 4 )
			cClrBits = 4;
		else if ( cClrBits <= 8 )
			cClrBits = 8;
		else if ( cClrBits <= 16 )
			cClrBits = 16;
		else if ( cClrBits <= 24 )
			cClrBits = 24;
		else
			cClrBits = 32;

		PBITMAPINFO		pbi;
		if (cClrBits != 24)
		{
			pbi = (PBITMAPINFO) ::LocalAlloc(LPTR, sizeof(BITMAPINFOHEADER) +
				sizeof(RGBQUAD) * (1<< cClrBits));
		}
		else
			pbi = (PBITMAPINFO) ::LocalAlloc(LPTR, sizeof(BITMAPINFOHEADER));

		pbi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		pbi->bmiHeader.biWidth = bmp.bmWidth;
		pbi->bmiHeader.biHeight = bmp.bmHeight;
		pbi->bmiHeader.biPlanes = bmp.bmPlanes;
		pbi->bmiHeader.biBitCount = bmp.bmBitsPixel;
		if (cClrBits < 24)
			pbi->bmiHeader.biClrUsed = (1<<cClrBits);

		pbi->bmiHeader.biCompression = BI_RGB;
		pbi->bmiHeader.biSizeImage = ((pbi->bmiHeader.biWidth * cClrBits +31) & ~31) /8
			* pbi->bmiHeader.biHeight;
		pbi->bmiHeader.biClrImportant = 0;

		PBITMAPINFOHEADER		pbih = (PBITMAPINFOHEADER) pbi;
		LPBYTE					lpBits = (LPBYTE) ::GlobalAlloc( GMEM_FIXED, pbih->biSizeImage );
		if ( lpBits )
		{
			if ( ::GetDIBits( hDC, hBitmap, 0, (WORD) pbih->biHeight, lpBits, pbi, DIB_RGB_COLORS ) )
			{
				BITMAPFILEHEADER		hdr;
				hdr.bfType = ((WORD) ('M' << 8) | 'B');
				hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) +
					pbih->biSize + pbih->biClrUsed *
					sizeof(RGBQUAD) + pbih->biSizeImage);
				hdr.bfReserved1 = 0;
				hdr.bfReserved2 = 0;
				hdr.bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER) +
					pbih->biSize + pbih->biClrUsed *
					sizeof (RGBQUAD);

				size_t			sBMP = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) +
					pbih->biClrUsed * sizeof (RGBQUAD) + pbih->biSizeImage;
				BYTE*			pbBMP = (BYTE*) ::malloc( sBMP );
				if ( pbBMP )
				{
					*( (BITMAPFILEHEADER*)pbBMP ) = hdr;
					*( (BITMAPINFOHEADER*)( pbBMP + sizeof( BITMAPFILEHEADER ) ) ) = *pbih;
					::memcpy( pbBMP + sizeof( BITMAPFILEHEADER ) + sizeof( BITMAPINFOHEADER ),
						(BYTE*) pbih + sizeof( BITMAPINFOHEADER ),
						pbih->biClrUsed * sizeof (RGBQUAD) );
					::memcpy( pbBMP + sizeof( BITMAPFILEHEADER ) + sizeof( BITMAPINFOHEADER ) +
						pbih->biClrUsed * sizeof (RGBQUAD),
						lpBits, pbih->biSizeImage );

					SResponseCharacteristics		srchrs;
					ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
					srchrs.pszContentType = "image/bmp";
					srchrs.uiContentLength = sBMP;
					srchrs.pbData = pbBMP;
					sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

					::free( pbBMP );
				}
			}

			::GlobalFree( (HGLOBAL)lpBits );
		}

		::LocalFree( pbi );
	}

	::SelectObject( hDC, hOldBitmap );
	::DeleteDC( hDC );
	::DeleteObject( hBitmap );

	// free the icon

	::DestroyIcon( shfiInfo.hIcon );

	// return to the caller
	return sResponseLen;
}

//
// Return_RightFrameStart --> return the default html for the right frame
//

#define ___CRLF "\r\n"

size_t Return_RightFrameStart( UINT uiProtocolVersion, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen )
{
	size_t		sResponseLen = 0;

	// compose the html text for the page

	char		szHtml[ 16 * 1024 ] = "";

	::strcat( szHtml, "<HTML>" ___CRLF );
	::strcat( szHtml, "<HEAD>" ___CRLF );
	::strcat( szHtml, "  <TITLE>Essential WWW Server Example</TITLE>" ___CRLF );
	::strcat( szHtml, "  <STYLE>" ___CRLF );
	::strcat( szHtml, "  BODY" ___CRLF );
	::strcat( szHtml, "  {" ___CRLF );
	::strcat( szHtml, "    font-family:verdana,arial,helvetica;" ___CRLF );
	::strcat( szHtml, "    margin:0;" ___CRLF );
	::strcat( szHtml, "  }" ___CRLF );
	::strcat( szHtml, "  </STYLE>" ___CRLF );
	::strcat( szHtml, "</HEAD>" ___CRLF );
	::strcat( szHtml, "" ___CRLF );
	::strcat( szHtml, "<BODY SCROLL=\"YES\" RIGHTMARGIN=\"0\" BOTTOMMARGIN=\"0\" LEFTMARGIN=\"0\" TOPMARGIN=\"0\" MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\">" ___CRLF );
	::strcat( szHtml, "<TABLE WIDTH=\"100%\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF );
	::strcat( szHtml, "<TR BGCOLOR=\"#000000\">" ___CRLF );
	::strcat( szHtml, "  <TD ALIGN=\"LEFT\">" ___CRLF );
	::strcat( szHtml, "    <IMG SRC=\"/default/gif/BlackBox.gif\" WIDTH=\"352\" HEIGHT=\"264\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "  <TD ALIGN=\"RIGHT\" VALIGN=\"BOTTOM\">" ___CRLF );
	::strcat( szHtml, "    <IMG SRC=\"/default/gif/VpcLogo.gif\" WIDTH=\"108\" HEIGHT=\"50\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "<TR>" ___CRLF );
	::strcat( szHtml, "  <TD COLSPAN=\"2\" NOWRAP HEIGHT=\"4\" BACKGROUND=\"/default/gif/Fade.gif\"></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "<TR>" ___CRLF );
	::strcat( szHtml, "  <TD COLSPAN=\"2\"><BR></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "<TR BGCOLOR=\"#980000\">" ___CRLF );
	::strcat( szHtml, "  <TD COLSPAN=\"2\"><CENTER><FONT COLOR=\"#FFFFFF\" SIZE=\"2\"><B><BR>Essential WWW Server Extension Example<BR>Copyright(c) 2002, VPC Technologies<BR><BR></B></FONT></CENTER></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "</TABLE>" ___CRLF );

	::strcat( szHtml, "</BODY>" ___CRLF );
	::strcat( szHtml, "" ___CRLF );
	::strcat( szHtml, "</HTML>" ___CRLF );

	// compose the response

	SResponseCharacteristics		srchrs;
	ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
	srchrs.pszContentType = "text/html";
	srchrs.uiContentLength = ::strlen( szHtml );
	srchrs.pbData = (BYTE*) szHtml;
	sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

	// return to the caller
	return sResponseLen;
}

#undef ___CRLF

//
// Return_LocalDrivesStart --> return the html for the "local drives" page
//

#define ___CRLF "\r\n"

size_t Return_LocalDrivesStart( UINT uiProtocolVersion, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen )
{
	size_t		sResponseLen = 0;

	// compose the html text for the page

	char		szHtml[ 16 * 1024 ] = "";

	::strcat( szHtml, "<HTML>" ___CRLF );
	::strcat( szHtml, "<HEAD>" ___CRLF );
	::strcat( szHtml, "  <TITLE>Essential WWW Server Example</TITLE>" ___CRLF );
	::strcat( szHtml, "  <STYLE>" ___CRLF );
	::strcat( szHtml, "  BODY" ___CRLF );
	::strcat( szHtml, "  {" ___CRLF );
	::strcat( szHtml, "    font-family:verdana,arial,helvetica;" ___CRLF );
	::strcat( szHtml, "    margin:0;" ___CRLF );
	::strcat( szHtml, "  }" ___CRLF );
	::strcat( szHtml, "  </STYLE>" ___CRLF );
	::strcat( szHtml, "</HEAD>" ___CRLF );
	::strcat( szHtml, "" ___CRLF );
	::strcat( szHtml, "<BODY SCROLL=\"YES\" RIGHTMARGIN=\"0\" BOTTOMMARGIN=\"0\" LEFTMARGIN=\"0\" TOPMARGIN=\"0\" MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\">" ___CRLF );
	::strcat( szHtml, "<TABLE WIDTH=\"100%\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF );
	::strcat( szHtml, "<TR BGCOLOR=\"#F1F1F1\">" ___CRLF );
	::strcat( szHtml, "  <TD WIDTH=\"100%\" HEIGHT=\"7\" COLSPAN=\"5\"></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "<TR BGCOLOR=\"#F1F1F1\">" ___CRLF );
	::strcat( szHtml, "  <TD>&nbsp;&nbsp;</TD>" ___CRLF );
	::strcat( szHtml, "  <TD VALIGN=\"MIDDLE\">" ___CRLF );
	::strcat( szHtml, "    <TABLE WIDTH=\"91\" HEIGHT=\"91\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF );
	::strcat( szHtml, "      <TR>" ___CRLF );
	::strcat( szHtml, "        <TD COLSPAN=\"3\">" ___CRLF );
	::strcat( szHtml, "          <IMG SRC=\"/default/gif/CircleUp.gif\" WIDTH=\"91\" HEIGHT=\"29\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "      </TR>" ___CRLF );
	::strcat( szHtml, "      <TR>" ___CRLF );
	::strcat( szHtml, "        <TD>" ___CRLF );
	::strcat( szHtml, "          <IMG SRC=\"/default/gif/CircleLeft.gif\" WIDTH=\"29\" HEIGHT=\"32\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "        <TD>" ___CRLF );
	::strcat( szHtml, "          <IMG SRC=\"/default/gif/LocalDrivesStart.gif\" WIDTH=\"32\" HEIGHT=\"32\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "        <TD>" ___CRLF );
	::strcat( szHtml, "          <IMG SRC=\"/default/gif/CircleRight.gif\" WIDTH=\"30\" HEIGHT=\"32\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "      </TR>" ___CRLF );
	::strcat( szHtml, "      <TR>" ___CRLF );
	::strcat( szHtml, "        <TD COLSPAN=\"3\">" ___CRLF );
	::strcat( szHtml, "          <IMG SRC=\"/default/gif/CircleDown.gif\" WIDTH=\"91\" HEIGHT=\"30\" BORDER=\"0\" /></TD>" ___CRLF );
	::strcat( szHtml, "      </TR>" ___CRLF );
	::strcat( szHtml, "    </TABLE></TD>" ___CRLF );
	::strcat( szHtml, "  <TD>&nbsp;&nbsp;</TD>" ___CRLF );
	::strcat( szHtml, "  <TD ALIGN=\"CENTER\" VALIGN=\"MIDDLE\" WIDTH=\"100%\">" ___CRLF );
	::strcat( szHtml, "    <FONT COLOR=\"#000000\" SIZE=\"6\">" ___CRLF );
	::strcat( szHtml, "      Local Drives" ___CRLF );
	::strcat( szHtml, "    </FONT></TD>" ___CRLF );
	::strcat( szHtml, "  <TD>&nbsp;&nbsp;</TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "<TR BGCOLOR=\"#F1F1F1\">" ___CRLF );
	::strcat( szHtml, "  <TD WIDTH=\"100%\" HEIGHT=\"7\" COLSPAN=\"5\"></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "<TR BGCOLOR=\"#D2D2D2\">" ___CRLF );
	::strcat( szHtml, "  <TD WIDTH=\"100%\" HEIGHT=\"1\" COLSPAN=\"5\"></TD>" ___CRLF );
	::strcat( szHtml, "</TR>" ___CRLF );
	::strcat( szHtml, "</TABLE>" ___CRLF );

	::strcat( szHtml, "<BR><BR><BR><CENTER><IMG SRC=\"/default/gif/TreeView.gif\" WIDTH=\"32\" HEIGHT=\"32\" BORDER=\"0\" /></CENTER>" ___CRLF );
	::strcat( szHtml, "<BR><BR><CENTER><FONT COLOR=\"#000000\" SIZE=\"2\"><B>Select a file or a directory to edit from the TreeView at the left.<BR><BR>Click <A STYLE=\"CURSOR:hand\" ONCLICK=\"UpdateTreeView();\"><FONT COLOR=\"#FF0000\">here</FONT></A> to update the \"Local Drives\" node.</B></FONT></CENTER>" ___CRLF );

	::strcat( szHtml, "<SCRIPT LANGUAGE=\"JAVASCRIPT\">" ___CRLF );
	::strcat( szHtml, "function UpdateTreeView()" ___CRLF );
	::strcat( szHtml, "{" ___CRLF );
	::strcat( szHtml, "  try" ___CRLF );
	::strcat( szHtml, "  {" ___CRLF );
	::strcat( szHtml, "    top.deeptree.UpdateTreeNode(\"local_drives\");" ___CRLF );
	::strcat( szHtml, "  }" ___CRLF );
	::strcat( szHtml, "  catch(e)" ___CRLF );
	::strcat( szHtml, "  {" ___CRLF );
	::strcat( szHtml, "  }" ___CRLF );
	::strcat( szHtml, "}" ___CRLF );
	::strcat( szHtml, "</SCRIPT>" ___CRLF );

	::strcat( szHtml, "</BODY>" ___CRLF );
	::strcat( szHtml, "" ___CRLF );
	::strcat( szHtml, "</HTML>" ___CRLF );

	// compose the response

	SResponseCharacteristics		srchrs;
	ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
	srchrs.pszContentType = "text/html";
	srchrs.uiContentLength = ::strlen( szHtml );
	srchrs.pbData = (BYTE*) szHtml;
	sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

	// return to the caller
	return sResponseLen;
}

#undef ___CRLF

//
// Helper_ReturnHtmlStringRepr --> return the html string representation of csString
//

static int viChrImgsWidths[] = {
	/* from 32 */ 9, 5, 9, 15, 12, 21, 14, 5,
	8, 8, 11, 12, 7, 8, 5, 12, 12, 9,
	12, 12, 13, 12, 12, 12, 12, 12, 5, 7,
	13, 13, 13, 11, 17, 16, 13, 13, 14, 12,
	11, 14, 14, 9, 10, 13, 11, 16, 14, 14,
	13, 14, 14, 12, 13, 14, 15, 20, 15, 15,
	12, 8, 12, 8, 13, 13, 6, 12, 12, 10,
	12, 12, 9, 12, 12, 5, 8, 12, 5, 17,
	12, 12, 12, 12, 9, 10, 9, 12, 13, 16,
	12, 13, 11, 12, 5, 12, 15, 5, 13, 12,
	12, 12, 12, 12, 12, 10, 12, 12, 12, 9,
	7, 6, 16, 16, 12, 17, 19, 12, 12, 12,
	12, 12, 13, 14, 14, 12, 11, 14, 12, 13,
	12, 6, 12, 12, 12, 14, 10, 10, 11, 17,
	13, 18, 18, 5, 11, 11, 5, 5, 5, 5,
	5, 16, 16, 16, 17, 5, 5, 12, 12, 11,
	15, 12, 12, 8, 8, 12, 8, 12, 12, 16,
	12, 12, 8, 8, 5, 8, 12, 11, 12, 15,
	12, 12, 12, 5, 9, 9, 9, 12, 12, 5,
	13, 5, 9, 13, 14, 12, 14, 14, 12, 14,
	12, 12, 13, 14, 14, 14, 13, 15, 13, 6,
	8, 12, 13, 19, 12, 12, 13, 7, 9, 9,
	9, 9, 9, 9, 5, 9
};

charstring Helper_ReturnHtmlStringRepr( charstring& csString )
{
	charstring		csRetVal;

	// do the requested operation

	char		szBuffer[ 1024 ];

	for ( int i=0; i<csString.size(); i++ )
	{
		unsigned char		c = csString[ i ];
		if ( c < 32 )
			c = 32;

		if ( c != 32 )
			::sprintf( szBuffer, "<IMG SRC=\"/default/abc/%i.gif\" WIDTH=\"%i\" HEIGHT=\"20\" />", c, viChrImgsWidths[ c - 32 ] );
		else
			::strcpy( szBuffer, " " );
		csRetVal += szBuffer;
	}

	// return to the caller

	return csRetVal;
}

//
// Return_FilePage --> return the html that represents the "file page" for the specified file
//

#define APPEND csHtml += szBuffer;
#define ___CRLF "\r\n"

size_t Return_FilePage( UINT uiProtocolVersion, char* pszFileName, SServerFunctions& ssf, BYTE* pbResponse, size_t sResponseBufferLen, BYTE* pbRequest, size_t sRequestLen )
{
	size_t					sResponseLen = 0;

	// decode the file name

	WCHAR		szFileName[ MAX_PATH ];
	if ( ( ::strlen( pszFileName ) / 4 ) * 3 >= sizeof( szFileName ) - sizeof( WCHAR ) )
		return 0;
	Base64Encoding::decode( (BYTE*) szFileName, pszFileName, sizeof( WCHAR ) );

	// compose the html text for the page

	charstring	csHtml;
	char		szBuffer[ 16*1024 ];

	char		szPathString[ MAX_PATH * 2 ];
	char		szAnsiFullPathName[ MAX_PATH ];
	if ( ::WideCharToMultiByte( 850, 0, szFileName, -1, szAnsiFullPathName, sizeof( szAnsiFullPathName ), NULL, NULL ) == 0 )
		::WideCharToMultiByte( CP_ACP, 0, szFileName, -1, szAnsiFullPathName, sizeof( szAnsiFullPathName ), NULL, NULL );

	size_t		sAnsiFullPathNameLen = ::strlen( szAnsiFullPathName );

	BOOL		bIsFile = FALSE;
	BOOL		bIsDirectory = FALSE;

	if ( sAnsiFullPathNameLen != 0 &&
		( szAnsiFullPathName[ sAnsiFullPathNameLen - 1 ] == '\\' ||
		szAnsiFullPathName[ sAnsiFullPathNameLen - 1 ] == '/' ) )
	{
		if ( sAnsiFullPathNameLen > 3 )
			szAnsiFullPathName[ sAnsiFullPathNameLen - 1 ] = '\0';
		::strcpy( szPathString, szAnsiFullPathName );
		bIsDirectory = TRUE;
	}
	else
	{
		::strcpy( szPathString, szAnsiFullPathName );
		bIsFile = TRUE;
	}

	// compose the description.

	charstring		csDescription = "<FONT COLOR=\"#000000\" SIZE=\"1\">"
		"&nbsp;Object Type : ";

	if ( bIsDirectory )
	{
		csDescription += "Local Directory";
	}
	else if ( bIsFile )
	{
		csDescription += "Local File";
	}

	csDescription += "</FONT>";

	// compose the html stream to return to the client

	char		szIcon[ MAX_PATH * 5 ];
	::sprintf( szIcon, "/default/MsdnTreeXmlGen.dyn?Icon%s", pszFileName );

	csHtml += "<HTML>" ___CRLF;
	csHtml += "<HEAD>" ___CRLF;
	csHtml += "  <TITLE>Essential WWW Server Example</TITLE>" ___CRLF;
	csHtml += "  <STYLE>" ___CRLF;
	csHtml += "  BODY" ___CRLF;
	csHtml += "  {" ___CRLF;
	csHtml += "    font-family:verdana,arial,helvetica;" ___CRLF;
	csHtml += "    margin:0;" ___CRLF;
	csHtml += "  }" ___CRLF;
	csHtml += "  </STYLE>" ___CRLF;
	csHtml += "  <STYLE TYPE=\"text/css\">" ___CRLF;
	csHtml += "    A:link {color:\"#0000ff\";text-decoration:\"none\"}" ___CRLF;
	csHtml += "    A:visited {color:\"#0000ff\";text-decoration:\"none\"}" ___CRLF;
	csHtml += "    A:hover {color:\"#ff0000\";text-decoration:\"none\"}" ___CRLF;
	csHtml += "  </STYLE>" ___CRLF;
	csHtml += "</HEAD>" ___CRLF;
	csHtml += "" ___CRLF;
	csHtml += "<BODY SCROLL=\"YES\" RIGHTMARGIN=\"0\" BOTTOMMARGIN=\"0\" LEFTMARGIN=\"0\" TOPMARGIN=\"0\" MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\">" ___CRLF;
	csHtml += "<TABLE WIDTH=\"100%\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF;
	csHtml += "<TR BGCOLOR=\"#F1F1F1\">" ___CRLF;
	csHtml += "  <TD WIDTH=\"100%\" HEIGHT=\"7\" COLSPAN=\"5\"></TD>" ___CRLF;
	csHtml += "</TR>" ___CRLF;
	csHtml += "<TR BGCOLOR=\"#F1F1F1\">" ___CRLF;
	csHtml += "  <TD>&nbsp;&nbsp;</TD>" ___CRLF;
	csHtml += "  <TD VALIGN=\"MIDDLE\">" ___CRLF;
	csHtml += "    <TABLE WIDTH=\"91\" HEIGHT=\"91\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF;
	csHtml += "      <TR>" ___CRLF;
	csHtml += "        <TD COLSPAN=\"3\">" ___CRLF;
	csHtml += "          <IMG SRC=\"/default/gif/CircleUp.gif\" WIDTH=\"91\" HEIGHT=\"29\" BORDER=\"0\" /></TD>" ___CRLF;
	csHtml += "      </TR>" ___CRLF;
	csHtml += "      <TR>" ___CRLF;
	csHtml += "        <TD>" ___CRLF;
	csHtml += "          <IMG SRC=\"/default/gif/CircleLeft.gif\" WIDTH=\"29\" HEIGHT=\"32\" BORDER=\"0\" /></TD>" ___CRLF;
	csHtml += "        <TD>" ___CRLF;
	::sprintf( szBuffer, "          <IMG SRC=\"%s\" WIDTH=\"32\" HEIGHT=\"32\" BORDER=\"0\" /></TD>" ___CRLF, szIcon ); APPEND
	csHtml += "        <TD>" ___CRLF;
	csHtml += "          <IMG SRC=\"/default/gif/CircleRight.gif\" WIDTH=\"30\" HEIGHT=\"32\" BORDER=\"0\" /></TD>" ___CRLF;
	csHtml += "      </TR>" ___CRLF;
	csHtml += "      <TR>" ___CRLF;
	csHtml += "        <TD COLSPAN=\"3\">" ___CRLF;
	csHtml += "          <IMG SRC=\"/default/gif/CircleDown.gif\" WIDTH=\"91\" HEIGHT=\"30\" BORDER=\"0\" /></TD>" ___CRLF;
	csHtml += "      </TR>" ___CRLF;
	csHtml += "    </TABLE></TD>" ___CRLF;
	csHtml += "  <TD>&nbsp;&nbsp;</TD>" ___CRLF;
	csHtml += "  <TD WIDTH=\"100%\" HEIGHT=\"100%\">" ___CRLF;
	csHtml += "    <TABLE WIDTH=\"100%\" HEIGHT=\"100%\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF;
	csHtml += "      <TR HEIGHT=\"100%\">" ___CRLF;
	csHtml += "        <TD VALIGN=\"MIDDLE\" WIDTH=\"100%\" HEIGHT=\"100%\">" ___CRLF;
	csHtml += "          <CENTER>" + ::Helper_ReturnHtmlStringRepr( charstring( szPathString ) ) + "</CENTER></TD>" ___CRLF;
	csHtml += "      </TR>" ___CRLF;
	csHtml += "    </TABLE></TD>" ___CRLF;
	csHtml += "  <TD>&nbsp;&nbsp;</TD>" ___CRLF;
	csHtml += "</TR>" ___CRLF;
	csHtml += "<TR BGCOLOR=\"#F1F1F1\">" ___CRLF;
	csHtml += "  <TD WIDTH=\"100%\" HEIGHT=\"7\" COLSPAN=\"5\"></TD>" ___CRLF;
	csHtml += "</TR>" ___CRLF;
	csHtml += "<TR BGCOLOR=\"#D2D2D2\">" ___CRLF;
	csHtml += "  <TD WIDTH=\"100%\" HEIGHT=\"1\" COLSPAN=\"5\"></TD>" ___CRLF;
	csHtml += "</TR>" ___CRLF;
	csHtml += "</TABLE>" ___CRLF;

	csHtml += "<TABLE WIDTH=\"100%\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">" ___CRLF;
	csHtml += "  <TR BGCOLOR=\"#F0F0B0\"><TD>" + csDescription + "</TD></TR>" ___CRLF;
	csHtml += "  <TR BGCOLOR=\"#D2D2D2\"><TD WIDTH=\"100%\" HEIGHT=\"1\"></TD></TR>" ___CRLF;
	csHtml += "</TABLE>" ___CRLF;

	csHtml += "</BODY>" ___CRLF;
	csHtml += "" ___CRLF;
	csHtml += "</HTML>" ___CRLF;

	// compose the response

	SResponseCharacteristics		srchrs;
	ssf.pfnInitializeResponseCharacteristics( uiProtocolVersion, srchrs );
	srchrs.pszContentType = "text/html";
	srchrs.uiContentLength = csHtml.size();
	srchrs.pbData = (BYTE*) csHtml.c_str();
	sResponseLen = ssf.pfnReturn_Response( srchrs, pbResponse, sResponseBufferLen );

	// return to the caller
	return sResponseLen;
}

#undef ___CRLF
#undef APPEND

//
// exported ProcessHttpRequest: called in a multithreaded fashion when processing a request
//

extern "C" __declspec(dllexport) size_t _cdecl ProcessHttpRequest(
	UINT uiProtocolVersion,
	SServerFunctions& ssf,
	BYTE* pbRequest,
	size_t sRequestLen,
	BYTE* pbResponse,
	size_t sResponseBufferLen,
	unsigned long ulLocalAddr,
	unsigned long ulRemoteAddr
)
{
	size_t		sResponseLen = 0;

	// extract the query string

	char		szQueryString[ HTTPREQ_URI_MAXLEN ] = "";
	int			iPos = -1;

	for ( unsigned int i=0; i<sRequestLen-1; i++ )
		if ( pbRequest[i] == '\r' && pbRequest[i+1] == '\n' )
			break;
		else if ( pbRequest[i] == '?' && iPos == -1 )
			iPos = 0;
		else if ( pbRequest[i] == ' ' && iPos != -1 )
			break;
		else if ( iPos != -1 )
			szQueryString[ iPos++ ] = pbRequest[i];

	if ( iPos != -1 ) szQueryString[ iPos ] = '\0';

	// return the correct xml to the client

	if ( ::strlen( szQueryString ) == 0 )
	{
		// return the default xml for the root
		sResponseLen = Return_RootXml( uiProtocolVersion, ssf, pbResponse, sResponseBufferLen );
	}
	else if ( ::stricmp( szQueryString, "MyComputer" ) == 0 )
	{
		// return the xml for the root node
		sResponseLen = Return_RootNodeXml( uiProtocolVersion, ssf, pbResponse, sResponseBufferLen );
	}
	else if ( ::stricmp( szQueryString, "LocalDrives" ) == 0 )
	{
		// return the xml for the local drives
		sResponseLen = Return_LocalDrives( uiProtocolVersion, ssf, pbResponse, sResponseBufferLen );
	}
	else if ( ::memicmp( szQueryString, "Browse", 6 ) == 0 )
	{
		// return the xml for the requested folder
		sResponseLen = Return_Folder( uiProtocolVersion, szQueryString + 6, ssf, pbResponse, sResponseBufferLen );
	}
	else if ( ::memicmp( szQueryString, "Icon", 4 ) == 0 )
	{
		// return the icon for the requested file
		sResponseLen = Return_Icon( uiProtocolVersion, szQueryString + 4, ssf, pbResponse, sResponseBufferLen, FALSE );
	}
	else if ( ::memicmp( szQueryString, "FilePage", 8 ) == 0 )
	{
		// return the html that represents the "file page" for the requested file
		sResponseLen = Return_FilePage( uiProtocolVersion, szQueryString + 8, ssf, pbResponse, sResponseBufferLen, pbRequest, sRequestLen );
	}
	else if ( ::stricmp( szQueryString, "RightFrameStart" ) == 0 )
	{
		// return the default html for the right frame
		sResponseLen = Return_RightFrameStart( uiProtocolVersion, ssf, pbResponse, sResponseBufferLen );
	}
	else if ( ::stricmp( szQueryString, "LocalDrivesStart" ) == 0 )
	{
		// return the html for the page
		sResponseLen = Return_LocalDrivesStart( uiProtocolVersion, ssf, pbResponse, sResponseBufferLen );
	}
	else if ( ::memicmp( szQueryString, "SmallIcon", 9 ) == 0 )
	{
		// return the small icon for the requested file
		sResponseLen = Return_Icon( uiProtocolVersion, szQueryString + 9, ssf, pbResponse, sResponseBufferLen, TRUE );
	}

	// return to the caller
	return sResponseLen;
}
